/*
    Copyright (c) 2002 Jens Tinz
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* This file defines the items of the radial menu and their layout.
 *
 * To use this, first call init and then use MZPieActions.layouts. You
 * should not store your own reference to MZPieActions.layouts since
 * it is replaced on preference changes. Do not forget to call shutdown
 * to unregister the preference observer.
 *
 * To add standard actions add them to simpleActions. To add
 * nonstandard actions edit init. To add or modify layouts edit
 * _updateLayouts.
 */

// namespace/module. This file defines no other toplevel functions or vars.
var MZPieActions = {
    // Mapping from item name to Action instance:
    items: null,
    // Mapping from layout name to array of actions.
    layouts: null
};

/* Map from itemname to action function.
 * The item name is used as label name, the item name with the first "_"
 * replaced with a "/" and ".png" appended as icon name.
 */
MZPieActions.simpleActions = {
    // navigation
    navigation_stop: function(context) {
        BrowserStop();
    },
    navigation_reload: function(context) {
        BrowserReload();
    },
    navigation_reload_skip: function(context) {
        BrowserReloadSkipCache();
    },
    navigation_up: function(context) {
        var url = content.document.URL;
        url = url.slice(0, url.slice(0, -1).lastIndexOf('/') + 1);
        loadURI(url);
    },
    navigation_firstpage: function(context) {
        // code adapted from EasyGestures
        getWebNavigation().gotoIndex(0);
    },
    navigation_lastpage: function(context) {
        // code adapted from EasyGestures
        var nav = getWebNavigation();
        nav.gotoIndex(nav.sessionHistory.count - 1);
    },
    navigation_enter_address: function(context) {
        openLocation();
    },

    // tabs
    tab_new: function(context) {
        BrowserOpenTab();
    },
    tab_duplicate: function(context) {
        openNewTabWith(content.location.href, null, null, null, false);
    },
    tab_next: function(context) {
        getBrowser().tabContainer.selectedIndex++;
    },
    tab_prev: function(context) {
        getBrowser().tabContainer.selectedIndex--;
    },
    tab_close: function(context) {
        BrowserCloseTabOrWindow();
    },
    tab_close_others: function(context) {
        var browser = getBrowser();
        browser.removeAllTabsBut(browser.selectedTab);
    },
    tab_undo_close: function(context) {
        undoCloseTab(0);
    },

    // windows
    window_new: function(context) {
        var url = '';
        var prefs = Components.classes[
            '@mozilla.org/preferences-service;1']
            .getService(Components.interfaces.nsIPrefService)
            .getBranch('browser.startup.');
        if (prefs.getIntPref('page') === 1)
            url = prefs.getCharPref('homepage');

        window.open(url);
    },
    window_duplicate: function(context) {
        window.open(content.location);
    },
    window_close: function(context) {
        // XXX this should probably trigger the same action file ->
        // close window uses, but need to check that does something
        // sane in fullscreen mode (check the ctrl+w keybinding?)
        closeWindow(true);
    },
    window_close_others: function(context) {
        // adapted from easyGestures
        var enumerator = Components.classes[
            '@mozilla.org/appshell/window-mediator;1']
            .getService(Components.interfaces.nsIWindowMediator)
            .getEnumerator('navigator:browser');
        while (enumerator.hasMoreElements()) {
            var awin = enumerator.getNext();
            if (awin !== window)
                awin.close();
        }
    },
    window_minimize: function(context) {
        window.minimize();
    },
    window_fullscreen: function(context) {
        BrowserFullScreen();
    },
    window_togglebar: function(context) {
        // code adapted from EasyGestures
        toggleSidebar('viewBookmarksSidebar');
    },

    // page
    page_bookmark: function(context) {
        // XXX this needs to be updated for Places (Firefox 3).
        addBookmarkAs(getBrowser(), false);
    },
    page_save: function(context) {
        saveDocument(content.document);
    },
    page_info: function(context) {
        BrowserPageInfo();
    },
    page_source: function(context) {
        if (context.selection) {
            try {
                gContextMenu.viewPartialSource('selection');
                return;
            } catch (e) {
                // Fall through to "no selection".
            }
        }
        BrowserViewSourceOfDocument(content.document);
    },
    page_validate: function(context) {
        loadURI('http://validator.w3.org/check?uri=' + content.document.URL);
    },

    // frame
    frame_view: function(context) {
        loadURI(context.frameDoc.URL);
    },
    frame_reload: function(context) {
        context.frameDoc.location.reload();
    },
    frame_save: function(context) {
        saveDocument(context.frameDoc);
    },
    frame_info: function(context) {
        BrowserPageInfo(context.frameDoc);
    },
    frame_source: function(context) {
        BrowserViewSourceOfDocument(context.frameDoc);
    },

    // link
    link_in_newtab: function(context) {
        // XXX this function's interface changed in Firefox 3. Need to
        // check this is still sane. Specifically I bypass the
        // "security check".

        // #^%*#@^$*()#@%&*) browser/base/content/tabbrowser.xml has a
        // $#&*^#&*( try { ... } catch ex) { } around the
        // loadURIWithFlags call to make sure debugging invalid values
        // is sufficiently painful (if you get the arguments wrong
        // here the new tab will open but the page will not load).
        // Like when you remove the makeURI() call below
        // (just pass the string).
        openNewTabWith(context.link.href, null, null, null, false,
                       makeURI(context.frameDoc.location.href));
    },
    link_in_newwin: function(context) {
        // XXX this function's interface changed in Firefox 3. Need to
        // check this is still sane. Specifically I bypass the
        // "security check".
        openNewWindowWith(context.link.href, null, null, false,
                          makeURI(context.frameDoc.location.href));
    },
    link_save: function(context) {
        // XXX how does innerHTML make any sense here?
        saveURL(context.link.href, context.link.innerHTML,
                null, false, false, makeURI(context.frameDoc.location.href));
    },
    link_copy_url: function(context) {
        Components.classes['@mozilla.org/widget/clipboardhelper;1']
            .getService(Components.interfaces.nsIClipboardHelper)
            .copyString(context.link.href);
    },
    link_copy_email: function(context) {
        Components.classes['@mozilla.org/widget/clipboardhelper;1']
            .getService(Components.interfaces.nsIClipboardHelper)
            .copyString(context.link.href.slice(7));
    },

    // image
    image_view: function(context) {
        loadURI(context.image.src);
    },
    image_copy_url: function(context) {
        Components.classes['@mozilla.org/widget/clipboardhelper;1']
            .getService(Components.interfaces.nsIClipboardHelper)
            .copyString(context.image.src);
    },
    image_save: function(context) {
        saveImageURL(context.image.src, content.document.title,
                     null, false, false,
                     makeURI(context.frameDoc.location.href));
    },

    // selection
    selection_copy: function(context) {
        Components.classes['@mozilla.org/widget/clipboardhelper;1']
            .getService(Components.interfaces.nsIClipboardHelper)
            .copyString(context.selection);
    },
    selection_open: function(context) {
        loadURI(context.selection);
    },
    selection_open_ntab: function(context) {
        openNewTabWith(context.selection, null, null, null, true);
    },
    selection_search: function(context) {
        loadURI('http://www.google.com/search?q=' + context.selection);
    },
    selection_search_ntab: function(context) {
        openNewTabWith(
            'http://www.google.com/search?q=' + context.selection,
            null, null, null, false);
    },
    selection_translate: function(context) {
        loadURI('http://babelfish.altavista.com/');
        // XXX setTimeout abuse, should probably use an onload event here.
        setTimeout(function(selection) {
            content.document.getElementsByTagName(
                'textarea')[0].value = selection;
        }, 3000, context.selection);
    },
    selection_translate_ntab: function(context) {
        openNewTabWith('http://babelfish.altavista.com/',
                       null, null, null, false);
        // XXX setTimeout abuse, should probably use an onload event here.
        setTimeout(function(selection) {
            content.document.getElementsByTagName(
                'textarea')[0].value = selection;
        }, 3000, context.selection);
    },

    // edit
    edit_cut: function(context) {
        goDoCommand('cmd_cut');
    },
    edit_copy: function(context) {
        goDoCommand('cmd_copy');
    },
    edit_paste: function(context) {
        goDoCommand('cmd_paste');
    },
    edit_undo: function(context) {
        goDoCommand('cmd_undo');
    }
};

// Items from the previous map with a nonstandard icon.
MZPieActions.iconRemap = {
    navigation_reload_skip: 'navigation_reload',
    tab_duplicate: 'tab_new',
    tab_close_others: 'tab_close',
    window_duplicate: 'window_new',
    window_close_others: 'window_close',
    window_togglebar: 'window_minimize',
    selection_open_ntab: 'selection_open',
    selection_search_ntab: 'selection_search',
    selection_translate_ntab: 'selection_translate'
};


/* Helper object used by the scrollMenu.
 *
 * This manages a timer (using the DOM's setInterval) which handles
 * scrolling in a single direction.
 *
 * Initialize with a scrollFunc which is called with a distance in
 * (integer) pixels.
 *
 * Background info on what this does and why:
 *
 * Mozilla forces the interval on timers to be >= 10 ms. On my windows
 * system it seems to be forced to >= 15 ms. Larger intervals give
 * more different speeds (speed changes smoother when the mouse
 * moves). It also seems that on average (but *only* on average) the
 * "sensitivity" is better than 15 ms (at least on the windows system
 * I am currently testing on): running a setInterval loop with an
 * interval of 20 or 21 seconds 100 times gives an average interval
 * within 0.5 ms of the target. The *individual* deltas may or may not
 * be any good, since I am measuring with Date.now() and I do not
 * trust its resolution. It claims the individual deltas vary from 2
 * to 50 ms.
 *
 * Some (limited) experimenting indicates using setTimeout or running
 * a setInterval once then cancelling and reinitializing it with the
 * same timeout value has noticable (a few ms) overhead. So try to
 * leave the setInterval running if possible.
 *
 * Scrolling (window.scrollBy) works in integer pixels. We definitely
 * need to scroll more than one pixel per interval for higher scroll
 * speeds, but to allow really slow scrolling while also keeping fast
 * scrolls smooth we *also* vary the scroll interval.
 *
 * To DTRT when we get a lot (several per scroll interval) of setSpeed
 * calls we cannot just cancel the current timeout in setSpeed and set
 * a new one because we do not know how far we are from the timeout
 * expiring. If we did that we would stop scrolling as long as the
 * mouse is moving. We cannot reliably track how far we are from the
 * timeout by hand because we cannot query the current time with a
 * reasonable resolution (Date.now has a resolution of about 15ms on
 * windows on at least some Firefox versions). So we should usually
 * wait for the current interval to trigger before setting a new one.
 * Because setInterval seems to be more accurate than setTimeout if we
 * run it more than once we do want to use setInterval instead of just
 * always using setTimeout and unconditionally rescheduling at the end
 * of the timeout function.
 *
 * Most of the time this is a nonissue because the next timeout will
 * happen "soon enough" (compared to human reaction times) anyway. But
 * if we are currently scrolling *really* slowly and want to speed up
 * we need to cancel the timer. Specifically currently
 * (minScrollInterval 30 ms, scrollSpeed 0.0002) our slowest scroll
 * speeds are 1 pixel in 5 *seconds*, then 1.25 seconds, ~0.6 seconds,
 * ~0.3 seconds, etc. If we are in that 5 second timeout cancelling it
 * is almost certainly an improvement. So we only reset our interval
 * if we are going from really slow to normal or fast scrolling. In
 * other cases we wait for the next timer event and reset from there.
 *
 * Currently the rule used is "if the current timeout is more than
 * twice the new timeout we cancel it" (because in that case it is
 * fairly likely our new timeout will happen sooner than the one we
 * are cancelling). Notice this implies that the previous scroll
 * distance is 1 pixel. The next interval is >= minScrollInterval, so
 * if currentInterval >= 2 * nextInterval then currentinterval >= 2 *
 * minScrollInterval. If we would scroll one more pixel per interval
 * the time between intervals would be currentInterval * (1 - 1 / d),
 * and if d >= 2 then (1 - 1 / d) >= 1/2, so that would still be >=
 * minScrollInterval, so we would be doing that. So we only cancel the
 * old timer if it was scrolling at a rate of 1 pixel and either the
 * new scroll rate is also one pixel (so even if we accidentally delay
 * the next scroll a lot it will not be that noticable) or it has a
 * timeout of < 2 * minScrollInterval (so it is fast enough that
 * rescheduling is almost certainly "safe", and we will not reschedule
 * again until this one runs).
 */

MZPieActions.Scroller = function(scrollFunc) {
    this.scrollFunc = scrollFunc;
    // Time in ms between scroll events. Values are meaningless if
    // this.timer === null.
    this.currentInterval = this.nextInterval = -1;
    // Distance in integer (!) pixels to scroll per event. Also
    // meaningless if this.timer === null.
    this.distance = 0;
    // Id from setInterval or null if we are paused.
    this.timer = null;
    // Dynamically adjusted minScrollInterval.
    this.minScrollInterval = this.defaultMinScrollInterval;
    this.startTime = null;
    this.loops = 0;
};

MZPieActions.Scroller.prototype = new Object;

/* The minimum delay between scroll events (ms per interval). Do not
 * lower this too much: if the browser cannot keep up this causes
 * awkward jumps in scroll speed when we switch from scrolling 1 pixel
 * per interval to scrolling 2 pixels per interval.
 */
MZPieActions.Scroller.prototype.defaultMinScrollInterval = 30;
/* The maximum delay between scroll events (ms per interval). If the
 * browser cannot keep up with this rate the scroll speed becomes
 * inaccurate. We need this to prevent system load or browser
 * "hickups" from causing insane scroll interval values, especially
 * since we currently cannot lower the min scroll interval.
 */
MZPieActions.Scroller.prototype.maxScrollInterval = 1000;

/* Update the speed in pixels / ms. Call this with 0 to stop scrolling.
 *
 * Calling with 0 cancels the running timer if there is one. You want
 * to do that before discarding this object.
 */
MZPieActions.Scroller.prototype.setSpeed = function(pixelsPerMs) {
    if (pixelsPerMs) {
        // We want our pixels per interval to be an integer value. We
        // also want to round it away from zero to make sure the
        // interval we calculate from this is > minScrollInterval.
        // That is *down* (floor instead of ceil) for numbers < 0:
        if (pixelsPerMs > 0)
            this.distance = Math.ceil(this.minScrollInterval * pixelsPerMs);
        else
            this.distance = Math.floor(this.minScrollInterval * pixelsPerMs);
        // We keep this a float in case setInterval does not round. If
        // setInterval does round that is no big deal (this number is
        // large enough that that does not matter much, and we cannot
        // be entirely precise in that case anyway).

        // Notice that distance has the same sign as pixelsPerMs, so
        // nextInterval is always positive.
        this.nextInterval = this.distance / pixelsPerMs;
        // Guess that waiting for the current timer to expire would
        // take too long if the current interval is more than twice
        // the new interval. Remember that we do not null out
        // currentInterval when we cancel the timer, so we need to
        // check for that here or we could end up passing null to
        // clearInterval, which could be bad.
        if (this.timer !== null &&
            this.currentInterval > 2 * this.nextInterval) {
            clearInterval(this.timer);
            this.timer = null;
        }
        // If we are currently paused or we just cancelled a long
        // timeout...
        if (this.timer === null) {
            // ... then start scrolling.
            this.currentInterval = this.nextInterval;
            this.timer = setInterval(this._runTimer, this.currentInterval,
                                     this);
            this.startTime = Date.now();
        }
    } else {
        // It is a request to stop scrolling.
        if (this.timer !== null) {
            clearInterval(this.timer);
            this.timer = null;
        }
    }
};

/* Callback function passed to setInterval.
 *
 * Cannot use "this", pass the instance as argument instead.
 */
MZPieActions.Scroller.prototype._runTimer = function(self)
{
    self.scrollFunc(self.distance);
    ++self.loops;
    // We must have had a few scroll events (to average out errors)
    // and we must have spent some time scrolling (because the
    // resolution of Date.now() is not that great)
    var expectedTime = self.loops * self.currentInterval;
    if (self.loops > 10 && expectedTime > 500) {
        var now = Date.now();
        var actualTime = now - self.startTime;
        if (actualTime > 1000) {
            if (actualTime > expectedTime * 1.1) {
                self.minScrollInterval = actualTime / self.loops;
                if (self.minScrollInterval > self.maxScrollInterval)
                    self.minScrollInterval = self.maxScrollInterval;
                // and reschedule the timer if necessary:
                if (self.minScrollInterval > self.nextInterval)
                    self.setSpeed(self.distance / self.nextInterval);
            }
            self.startTime = now;
            self.loops = 0;
        }
    }
    if (self.currentInterval !== self.nextInterval) {
        clearInterval(self.timer);
        self.currentInterval = self.nextInterval;
        self.timer = setInterval(self._runTimer, self.currentInterval, self);
        self.startTime = Date.now();
        self.loops = 0;
    }
};


/* Special menuitem triggering scrolling.
 *
 * This "freezes" the menu in place on the screen as an indication of
 * where to move the mouse to stop scrolling. That is unusual enough
 * that the implementation is rather "special".
 */
MZPieActions.ScrollMenu = function(label, layout) {
    MZPieMenu.Submenu.call(this, label, 'scroll/menu.png', layout);
    // These are set from activateMenu and killed when the menu closes.
    this.horizScroller = this.vertScroller = null;
    // Likewise.
    this.targetView = null;
};

MZPieActions.ScrollMenu.prototype = new MZPieMenu.Submenu;

/*
 * With the mouse d pixels over the edge of the menu, scroll speed is
 * d * d * scrollSpeed pixels per millisecond,
 * rounded to whole pixels per interval.
 * So the larger this value, the faster we scroll.
 */
MZPieActions.ScrollMenu.prototype.scrollSpeed = 0.000001;

MZPieActions.ScrollMenu.prototype.activateMenu = function(menu) {
    // Hide the labels even in "show always/immediately" mode.
    menu.hideLabels();

    // Define our scrollers.
    var targetView = menu.context.frameDoc.defaultView;
    this.horizScroller = new MZPieActions.Scroller(function (d) {
            targetView.scrollBy(d, 0);
        });
    this.vertScroller = new MZPieActions.Scroller(function (d) {
            targetView.scrollBy(0, d);
        });

    // Minor hack: mousemove uses this to disable the scrolling timer
    // when we cannot scroll in that direction.
    this.targetView = targetView;

    /* HACK: we completely hijack the PieMenu by replacing its
     * callback methods with ours. That is ugly, but guarantees we do
     * not install a misbehaving event handler if someone is doing
     * something unusual with the menu, and also takes care of
     * stopping the menu from moving around.
     */
    var self = this;
    menu.update = function(mx, my) { self.mousemove(menu, mx, my); };
    menu.showLabels = function() {};

    // A bit tricky since we need to call the original method.
    var close = menu.close;
    menu.close = function(doAction) {
        // Kill the timers and drop the reference.
        self.vertScroller.setSpeed(0);
        self.horizScroller.setSpeed(0);
        self.vertScroller = null;
        self.horizScroller = null;
        self.targetView = null;
        close.call(menu, doAction);
    };
};

MZPieActions.ScrollMenu.prototype.mousemove = function(menu, mx, my) {
    /* This used to calculate speed the same way the menu calculates
     * movement (with no scrolling while the mouse is in the hole in
     * the menu) but that feels odd in the common case of no
     * horizontal scrolling (moving the mouse horizontally affects
     * vertical scrolling speed). So now we simply calculate
     * horizontal and vertical speed separately.
     */

    // TODO experiment with different functions to get the speed from
    // the distance. The current distance-squared one is pretty random.

    // Possibly unnecessary optimization: we force the scroller off if
    // the corresponding scrollMax value is 0 (if we cannot scroll in
    // that direction). Done because I am not sure how well various
    // Moz versions deal with two timers with a short but nonidentical
    // interval (and it is very common to not have a horizontal
    // scrollbar).
    var d = mx - menu.x;
    var d2 = Math.abs(d) - menu.actionRadius;
    if (d2 > 0 && this.targetView.scrollMaxX)
        // Cursor outside the menu.
        this.horizScroller.setSpeed((d > 0 ? 1 : -1) *
                                    d2 * d2 * d2 * this.scrollSpeed);
    else
        this.horizScroller.setSpeed(0);

    d = my - menu.y;
    d2 = Math.abs(d) - menu.actionRadius;
    if (d2 > 0 && this.targetView.scrollMaxY)
        this.vertScroller.setSpeed((d > 0 ? 1 : -1) *
                                   d2 * d2 * d2 * this.scrollSpeed);
    else
        this.vertScroller.setSpeed(0);
};


/* Forward button that grays out when at the end of history. */
MZPieActions.NaviForward = function(label) {
    MZPieMenu.Action.call(this, label, 'navigation/forward.png');
};

MZPieActions.NaviForward.prototype = new MZPieMenu.Action;
MZPieActions.NaviForward.prototype.tweak = function(context) {
    try {
        this.active = getWebNavigation().canGoForward;
    } catch (e if e instanceof ReferenceError) {
        this.active = false;
    }
    if (this.active)
        this.iconpath = 'navigation/forward.png';
    else
        this.iconpath = 'navigation/forward_gray.png';
};

MZPieActions.NaviForward.prototype.doAction = function(context) {
    var webnav = getWebNavigation();
    if (webnav.canGoForward)
        webnav.goForward();
};


/* Back button that grays out when at the beginning of history. */
MZPieActions.NaviBack = function(label) {
    MZPieMenu.Action.call(this, label, 'navigation/back.png');
};

MZPieActions.NaviBack.prototype = new MZPieMenu.Action;
MZPieActions.NaviBack.prototype.tweak = function(context) {
    try {
        this.active = getWebNavigation().canGoBack;
    } catch (e if e instanceof ReferenceError) {
        this.active = false;
    }
    if (this.active)
        this.iconpath = 'navigation/back.png';
    else
        this.iconpath = 'navigation/back_gray.png';
};

MZPieActions.NaviBack.prototype.doAction = function(context) {
    var webnav = getWebNavigation();
    if (webnav.canGoBack)
        webnav.goBack();
};


/* Label displaying the server affected by image blocking. */
MZPieActions.ImageServerLabel = function(label) {
    MZPieMenu.Dummy.call(this);
    this.label = label;
};

MZPieActions.ImageServerLabel.prototype = new MZPieMenu.Dummy;
MZPieActions.ImageServerLabel.prototype.tweak = function(context) {
    if (context.image === null) {
        this.label = '';
        return;
    }
    var url = context.image.src;
    url = url.substring(url.indexOf('//') + 2, url.length);
    if (url.indexOf('www.') >= 0)
        url = url.substring(url.indexOf('www.') + 4,  url.length);
    url = url.substring(0, url.indexOf('/'));
    if (url.length > 14)
        url = url.substring(0, 12) + '...';

    this.label = '(' + url + ')';
};


/* Context-sensitive submenu. */
MZPieActions.ContextMenu = function(childmap) {
    // We initialize with dummy values, tweak() sets sane values.
    MZPieMenu.Submenu.call(this, '', '', []);
    this.map = childmap;
};

MZPieActions.ContextMenu.prototype = new MZPieMenu.Submenu;
MZPieActions.ContextMenu.prototype.tweak = function(context) {
    if (context.next === null)
        // No more contexts (currently unreached code because the
        // context chain ends in the scrollmenu which does not contain
        // a contextmenu child).
        return new MZPieMenu.Dummy();
    var next = this.map[context.next.layout];
    if (next === undefined)
        // We have no layout for this context. This is probably an error,
        // but just hide it for now.
        return new MZPieMenu.Dummy();
    // Pretend to be the next submenu. We cannot just return it because we
    // need to modify menu.context in activateMenu.
    this.label = next.label;
    this.iconpath = next.iconpath;
    this.layout = next.layout;
    return null;
};

MZPieActions.ContextMenu.prototype.activateMenu = function(menu) {
    // This is a bit dirty: relies on nothing keeping a second
    // reference to context. That should be a valid assumption though.
    menu.context = menu.context.next;
    /* But this is *very* dirty. This makes the scrollmenu work, which
     * is currently the only user of activateMenu. May have to
     * redesign this if activateMenu gains more users than us and the
     * scrollmenu.
     */
    this.map[menu.context.layout].activateMenu(menu);
};


/* Either the page or frame submenu depending on context. */
MZPieActions.ContextPage = function(label, framemenu, pagemenu) {
    // Submenu subclass to make the submenu arrow show up in the prefdialog.
    MZPieMenu.Submenu.call(this, label, 'page/menu.png', []);
    this.framemenu = framemenu;
    this.pagemenu = pagemenu;
    this.active = false;
};

MZPieActions.ContextPage.prototype = new MZPieMenu.Submenu;
MZPieActions.ContextPage.prototype.tweak = function(context) {
    if (!content)
        return null;
    else if (content.document !== context.frameDoc)
        return this.framemenu;
    else
        return this.pagemenu;
};


/* A stop button while loading, a reload button when done. */
MZPieActions.ContextStop = function(label, reloaditem, stopitem) {
    MZPieMenu.Dummy.call(this);
    this.label = label;
    this.reloaditem = reloaditem;
    this.stopitem = stopitem;
};

MZPieActions.ContextStop.prototype = new MZPieMenu.Dummy;
MZPieActions.ContextStop.prototype.tweak = function(context) {
    var stop = document.getElementById('Browser:Stop');
    if (stop === null || stop.hasAttribute('disabled'))
        return this.reloaditem;
    else
        return this.stopitem;
};


/* Either "block server" or "unblock server" depending on current state. */
MZPieActions.ContextBlockImg = function(label, unblocklabel, blocklabel) {
    MZPieMenu.Action.call(this, label, 'image/block.png');
    this.unblocklabel = unblocklabel;
    this.blocklabel = blocklabel;
    this.nsIPermissionManager = Components.interfaces.nsIPermissionManager;
    this.pmanager = Components.classes['@mozilla.org/permissionmanager;1']
        .getService(this.nsIPermissionManager);
};

MZPieActions.ContextBlockImg.prototype = new MZPieMenu.Action;
MZPieActions.ContextBlockImg.prototype.tweak = function(context) {
    if (context.image === null)
        return;
    var uri = Components.classes['@mozilla.org/network/standard-url;1']
        .createInstance(Components.interfaces.nsIURI);
    uri.spec = context.image.src;

    if (this.pmanager.testPermission(uri, 'image') ===
        this.nsIPermissionManager.DENY_ACTION) {
        this.label = this.unblocklabel;
        this.iconpath = 'image/unblock.png';
    } else {
        this.label = this.blocklabel;
        this.iconpath = 'image/block.png';
    }
};

MZPieActions.ContextBlockImg.prototype.doAction = function(menu) {
    var uri = Components.classes['@mozilla.org/network/standard-url;1']
        .createInstance(Components.interfaces.nsIURI);
    uri.spec = context.image.src;
    this.pmanager.add(uri, 'image',
                      (this.pmanager.testPermission(uri, 'image') ==
                       this.nsIPermissionManager.DENY_ACTION)
                      ? this.nsIPermissionManager.ALLOW_ACTION
                      : this.nsIPermissionManager.DENY_ACTION);
};


MZPieActions.submenuNames = [
    'selection',
    'image',
    'frame',
    'page',
    'window',
    'tab',
    'navigation',
];


/* Initialize the items and layouts mappings, set up preference observer.
 *
 * Call this once before accessing MZPieActions.layouts. Calling it more
 * than once is safe, as is calling it again after calling shutdown.
 *
 * rcbundle is the stringbundle containing label translations.
 * prefbranch is an nsIPrefBranch2 containing the "customizations" tree.
 */
MZPieActions.init = function(rcbundle, prefbranch) {
    /* We could easily create a prefbranch here instead of getting one
     * passed in, but then we would have to store a reference to it so
     * we can unregister the listener. Since the main user
     * (integration.js) already keeps a reference to the right
     * prefbranch anyway we do it this way.
     */
    if (this.items !== null)
        return;

    // Gettext-inspired shortcut.
    function _(label) {
        return rcbundle.getString(label);
    }

    var items = {};
    for (var [name, func] in Iterator(this.simpleActions)) {
        var icon = (this.iconRemap[name] || name).replace('_', '/') + '.png';
        var action = items[name] = new MZPieMenu.Action(_(name), icon);
        action.doAction = func;
    }

    // Dummies.
    items.dummy = new MZPieMenu.Dummy();
    items.image_server = new this.ImageServerLabel(_('image_server'));

    // Dummy actions.
    items.scroll_down = new MZPieMenu.Action(_('scroll_down'),
                                             'scroll/down.png');
    items.scroll_right = new MZPieMenu.Action(_('scroll_right'),
                                              'scroll/right.png');
    items.scroll_left = new MZPieMenu.Action(_('scroll_left'),
                                             'scroll/left.png');
    items.scroll_up = new MZPieMenu.Action(_('scroll_up'), 'scroll/up.png');

    // Actions.
    items.navigation_forward = new this.NaviForward(_('navigation_forward'));
    items.navigation_back = new this.NaviBack(_('navigation_back'));

    items.context_stop = new this.ContextStop(_('context_stop'),
                                              items.navigation_reload,
                                              items.navigation_stop);
    items.context_blockimg = new this.ContextBlockImg(_('context_blockimg'),
                                                      _('image_unblock'),
                                                      _('image_block'));

    // Submenus.

    /* HACK: instead of directly constructing the Submenu items (which
     * is nontrivial because of nested references, sometimes involving
     * context-sensitive items too) we pass an empty layout here, and
     * swap it for the actual layout later. This works because nothing
     * except for the actual PieMenu accesses Submenu.layout.
     */

    for each (var menu in this.submenuNames) {
        items['menu_' + menu] = new MZPieMenu.Submenu(_('menu_' + menu),
                                                      menu + '/menu.png', []);
    }

    items.context_page = new this.ContextPage(
        _('context_page'), items.menu_frame, items.menu_page);
    items.menu_scroll = new this.ScrollMenu(
        _('menu_scroll'), [
            items.dummy,            items.scroll_right,
            items.dummy,            items.scroll_down,
            items.dummy,            items.scroll_left,
            items.dummy,            items.scroll_up,
        ]);

    /* Triggered are "scroll" or any toplevel layout type. The only
     * supported nested node situation is currently image inside link,
     * other combinations will cause badness.
     */
    items.context_2 = new this.ContextMenu({
        scroll: items.menu_scroll,
        navigation: items.menu_navigation,
        image: items.menu_image,
        selection: items.menu_selection
    });

    // Alternate actions.
    items.navigation_reload.altitem = items.navigation_reload_skip;
    items.navigation_forward.altitem = items.navigation_lastpage;
    items.navigation_back.altitem = items.navigation_firstpage;
    items.navigation_up.altitem = items.navigation_enter_address;
    items.tab_new.altitem = items.tab_duplicate;
    items.tab_close.altitem = items.tab_close_others;
    items.window_new.altitem = items.window_duplicate;
    items.window_close.altitem = items.window_close_others;
    items.window_minimize.altitem = items.window_togglebar;
    items.page_info.altitem = items.page_validate;
    items.selection_open.altitem = items.selection_open_ntab;
    items.selection_search.altitem = items.selection_search_ntab;
    items.selection_translate.altitem = items.selection_translate_ntab;

    this.items = items;

    if (prefbranch !== null)
        prefbranch.addObserver('customizations.', this, false);

    this._updateLayouts(prefbranch);
};

/* Unregister pref observer. */
MZPieActions.shutdown = function(prefbranch) {
    prefbranch.removeObserver('customizations.', this);
    // Mainly to make calling init again after shutdown work.
    this.layouts = this.items = null;
};


/* Return [layoutName, index, value] or null for 'customizations.blah.2'.
 *
 * The prefbranch should contain the "customizations" key.
 *
 * value returned is null if unset and allowNullValue is true.
 */
MZPieActions.parseKey = function(prefbranch, key, allowNullValue) {
    var split = key.split('.');
    if (split.length !== 3)
        return null;
    // The first value is "customizations" which we ignore.
    var [, layoutName, index] = split;

    if (!this.layouts.hasOwnProperty(layoutName))
        return null;

    // Sanity check preventing multiple prefs specifing the same index:
    if (!/^(0|[1-9][0-9]*)$/.test(index))
        return null;
    index = parseInt(index);
    if (index >= this.layouts[layoutName].length)
        return null;

    // Ok, the key is valid.
    try {
        var value = prefbranch.getCharPref(key);
    } catch (e) {
        // Assume this failed because the pref is not set.
        if (allowNullValue)
            return [layoutName, index, null];
        else
            return null;
    }
    if (!this.items.hasOwnProperty(value))
        return null;

    // Value is valid too, store:
    return [layoutName, index, value];
};


/* Update MZPieActions.layouts. */
MZPieActions._updateLayouts = function(prefbranch) {
    /* The way this works with updates is hacky but effective. First,
     * updating the array in place works because nothing keeps a
     * longterm reference to a "layout" array: the menu copies it
     * (applying tweakfunctions) on initialization and when displaying
     * a submenu. And we have to do a full reinitialization on pref
     * change because we do get notified about the change but do not
     * get the previous value, so unless we manually track the applied
     * customizations we cannot apply just the changes.
     */
    var items = this.items;

    // Menu layouts.
    this.layouts = {
        // Primary layouts (the menu starts out in one of these):
        navigation: [
            items.menu_tab,         items.navigation_forward,
            items.context_page,     items.context_stop,
            items.context_2,        items.navigation_back,
            items.menu_window,      items.navigation_up,
        ],
        link: [
            items.menu_tab,         items.link_in_newtab,
            items.context_page,     items.link_save,
            items.context_2,        items.link_in_newwin,
            items.menu_window,      items.link_copy_url,
        ],
        mail: [
            items.menu_tab,         items.dummy,
            items.context_page,     items.dummy,
            items.context_2,        items.dummy,
            items.menu_window,      items.link_copy_email,
        ],
        edit: [
            items.menu_tab,         items.edit_copy,
            items.context_page,     items.edit_paste,
            items.context_2,        items.edit_cut,
            items.menu_window,      items.edit_undo,
        ],
        image: [
            items.menu_tab,         items.context_blockimg,
            items.context_page,     items.image_save,
            items.context_2,        items.image_view,
            items.menu_window,      items.image_copy_url,
        ],
        selection: [
            items.menu_tab,         items.selection_open,
            items.context_page,     items.selection_translate,
            items.context_2,        items.selection_search,
            items.menu_window,      items.selection_copy,
        ],
        // Submenus, opened through the context menu:
        navigation_sub: [
            items.context_2,        items.navigation_forward,
            items.dummy,            items.context_stop,
            items.dummy,            items.navigation_back,
            items.dummy,            items.navigation_up,
        ],
        image_sub: [
            items.context_2,        items.context_blockimg,
            items.image_server,     items.image_save,
            items.dummy,            items.image_view,
            items.dummy,            items.image_copy_url,
        ],
        selection_sub: [
            items.context_2,        items.selection_open,
            items.dummy,            items.selection_translate,
            items.dummy,            items.selection_search,
            items.dummy,            items.selection_copy,
        ],
        // Submenus, direct children of the main menu:
        tab_sub: [
            items.dummy,            items.tab_next,
            items.dummy,            items.tab_close,
            items.tab_undo_close,   items.tab_prev,
            items.dummy,            items.tab_new,
        ],
        window_sub: [
            items.dummy,            items.window_fullscreen,
            items.dummy,            items.window_close,
            items.dummy,            items.window_minimize,
            items.dummy,            items.window_new,
        ],
        // Either page or frame is also a child of the main menu:
        page_sub: [
            items.dummy,            items.page_bookmark,
            items.dummy,            items.page_save,
            items.dummy,            items.page_source,
            items.dummy,            items.page_info,
        ],
        frame_sub: [
            items.dummy,            items.frame_view,
            items.frame_reload,     items.frame_save,
            items.dummy,            items.frame_source,
            items.menu_page,        items.frame_info,
        ]
    };

    if (prefbranch !== null) {
        for each (var key in prefbranch.getChildList('customizations.', {})) {
            var entry = this.parseKey(prefbranch, key, false);
            if (entry !== null) {
                var [layoutName, index, value] = entry;
                this.layouts[layoutName][index] = items[value];
            }
        }
    }

    /* Now fix up the submenus as described above. */
    for each (var menu in this.submenuNames) {
        items['menu_' + menu].layout = this.layouts[menu + '_sub'];
    }
};


MZPieActions.observe = function(prefbranch, topic, data) {
    if (topic !== 'nsPref:changed')
        throw new Error('huh, I did not sign up for ' + topic);

    // This is required, without it the prefbranch only shows up as
    // nsISupports.
    prefbranch.QueryInterface(Components.interfaces.nsIPrefBranch);
    this._updateLayouts(prefbranch);
};
