/*
    Copyright (c) 2002 Jens Tinz
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* This file manages the integration of the radial menu with the browser.
 *
 * There's functionality for responding to events and reading preferences.
 */


/* Fake module scope.
 *
 * All data and functions are in here just to prevent name clashes.
 * But all functions in here are wrapped (see the bottom of this file)
 * so the functions in here can do this.blah instead of
 * MZPieIntegration.blah even when used as event handlers like
 * addEventListener('event', MZPieIntegration.handler, true);
 * while removeEventListener('event', this.handler, true) also still works.
 *
 * The point is you can rename the entire thing by renaming
 * MZPieIntegration in (currently) 5 places: here, the
 * addEventListener for init, the loop wrapping the functions (it'd be
 * 3 if that loop was in a helper function with the module passed as
 * arg). Useful for forking :)
 */

var MZPieIntegration = {

    // Constants: valid "enum" values for this.state.
    // No radial menu triggered:
    kStateHidden: 0,
    // Menu still invisible but mouse down (in "open on drag" mode):
    kStateOpenOnDrag: 1,
    // Menu was just drawn, mouse button is down:
    kStateShowing: 2,
    // Menu was dragged, mouse button is down:
    kStateDragged: 3,
    // Mouse button was released without moving the mouse, menu stays
    // open until any mouse button is clicked:
    kStateClickedOpen: 4,

    // our nsIPrefBranch:
    prefbranch: null,
    // a second prefbranch for the general dom contextmenu pref
    agressiveBranch: null,
    // Current state of the menu/mouse (see kState* for values):
    state: 0, // (cannot refer to this.kStateHidden here yet)
    // The "context" object passed to the menu.
    // Only used to store it between click and move in drag-open mode.
    context: null,
    // the node to add the menu to, same reason as context:
    parentNode: null,
    // Menu currently displayed. Invariant: is null if and only if
    // state is kStateHidden or kStateOpenOnDrag.
    menu: null,
    // Timeout for displaying the labels:
    labelTrigger: null,
    // The dom element we use for receiving mousemove events.
    mouseMoveElement: null,
    // Offset from screen to client coordinates
    screenToClientX: 0,
    screenToClientY: 0,
    // Whether we are preventing other scripts from interfering with the menu
    agressive: true,
    /* These two are used to work around two event handling glitches:
     * spurious mousemove events and omitted mousemove events.
     */
    lastMouseX: null,
    lastMouseY: null,

    /* Display the "tip of the day" dialog. */
    showTipDialog: function() {
        openDialog('chrome://radialcontext-mz/content/tipdialog.xul',
                   'rc_mz_tip', 'alwaysRaised,centerscreen');
    },

    /* Display the tooltip. */
    _showTooltip: function() {
        this.labelTrigger = null;
        this.menu.showLabels();
    },

    /* Reschedule the display of the tooltip. */
    _scheduleTooltip: function() {
        if (this.labelTrigger !== null)
            clearTimeout(this.labelTrigger);
        var tooltipDelay = this.prefbranch.getIntPref('ttip_delay');
        /* Tooltip behaviour for delay 0 and keep off is special.
         * Logically delay 0 should just force keep to true, but I
         * think the current behaviour might be useful too: it shows
         * the tooltip immediately and keeps it until the menu is
         * dragged. It would be even better if opening a submenu
         * redisplayed the tooltip (opening a submenu without moving
         * the parent menu already keeps the tooltip).
         *
         * Notice there is no check for an already displayed tooltip
         * here. showLabels is almost a noop in that case, so I ignore
         * that for now.
         */
        if (tooltipDelay > 0 && this.menu !== null)
            this.labelTrigger = setTimeout(this._showTooltip, tooltipDelay);
        else
            this.labelTrigger = null;
    },

    /* Initialize the menu. evt should be a mouse event, context an object. */
    _showMenu: function(evt, parentNode, context) {
        /* textZoom getting should be in menu.js, but I cannot find a
         * browser.js-independent way of doing it.
         *
         * getMarkupDocumentViewer is defined by browser.js as
         *   gBrowser.markupDocumentViewer
         * which is an xbl property on a "browser" object:
         *   this.docShell.contentViewer
         *       .QueryInterface(nsIMarkupDocumentViewer)
         * with docShell an xbl property:
         *   this.boxObject.QueryInterface(nsIBrowserBoxObject)
         *
         * nsIDOMWindow has a textZoom attribute, but it is marked
         * noscript (why not just "readonly" is beyond me).
         * nsIMarkupDocumentViewer and nsIDocShell have (text)zoom
         * attributes, but I see no clean way of getting them in the idl:
         * only nsIXULWindow, nsIBrowserBoxObject and nsIIFrameBoxObject
         * have docShell attrs. So I see no obvious clean way of getting
         * at the zoom value from just a document or window object, which
         * is why the menu cannot figure it out on its own.
         */
        this.menu = new MZPieMenu(
            parentNode, MZPieActions.layouts[context.layout],
            evt.screenX + this.screenToClientX,
            evt.screenY + this.screenToClientY, evt.altKey,
            this.prefbranch.getBoolPref('ttip_keep'),
            getMarkupDocumentViewer().textZoom, context);
        if (this.prefbranch.getIntPref('ttip_delay') === 0)
            this.menu.showLabels();
        this._scheduleTooltip();
        this.state = this.kStateShowing;
    },

    /* Hide the menu (assumes there is one).
     *
     * With doAction true also execute the selected action.
     */
    _hideMenu: function(doAction) {
        this._removeMouseMoveListener();
        if (this.labelTrigger !== null) {
            clearTimeout(this.labelTrigger);
            this.labelTrigger = null;
        }
        // The actual action functions are probably the most
        // frequently broken code, so make sure we deal if this
        // raises:
        try {
            this.menu.close(doAction);
        } finally {
            this.menu = null;
        }
    },

    /* Initialize members, hook up event handlers, display tip dialog. */
    init: function() {
        window.removeEventListener('load', this.init, true);

        var prefService = Components.classes
            ['@mozilla.org/preferences-service;1']
            .getService(Components.interfaces.nsIPrefService);
        this.prefbranch = prefService
            .getBranch('extensions.radialcontext-mz.')
            .QueryInterface(Components.interfaces.nsIPrefBranch2);

        this.agressiveBranch = prefService
            .getBranch('dom.event.contextmenu.')
            .QueryInterface(Components.interfaces.nsIPrefBranch2);

        this.agressiveBranch.addObserver('enabled', this, false);
        this.agressive = !this.agressiveBranch.getBoolPref('enabled');

        // show tip at startup
        if (this.prefbranch.getBoolPref('tips_at_startup')) {
            // count windows
            var wincnt = 0;
            var wenum = Components.classes[
                '@mozilla.org/embedcomp/window-watcher;1']
                .getService(Components.interfaces.nsIWindowWatcher)
                .getWindowEnumerator();
            while (wenum.hasMoreElements()) {
                wenum.getNext();
                ++wincnt;
            }
            // show dialog for the first window only
            if (wincnt === 1)
                this.showTipDialog();
        }

        // prepare menu items and register customization pref listener:
        MZPieActions.init(document.getElementById('radialcontext_mz_bundle'),
                          this.prefbranch);

        // add listeners
        window.addEventListener('unload', this.shutdown, false);
        this.addEventListeners(getBrowser());
    },

    addEventListeners: function(browser) {
        browser.addEventListener('mousedown', this.onMouseDown,
                                 this.agressive);
        browser.addEventListener('mouseup', this.onMouseUp, this.agressive);
        browser.addEventListener('keydown', this.onKeyDown, this.agressive);
        browser.addEventListener('keyup', this.onKeyUp, this.agressive);
        document.getElementById('contentAreaContextMenu').addEventListener(
            'popupshowing', this.onPopup, this.agressive);
        browser.addEventListener('contextmenu', this.onContextMenu,
                                 this.agressive);
    },

    _removeMouseMoveListener: function() {
        this.mouseMoveElement.removeEventListener(
            'mousemove', this.onMouseMove, true);
        this.mouseMoveElement = null;
        this.state = this.kStateHidden;
    },

    removeEventListeners: function(browser) {
        browser.removeEventListener('mousedown', this.onMouseDown,
                                 this.agressive);
        browser.removeEventListener('mouseup', this.onMouseUp,
                                    this.agressive);
        browser.removeEventListener('keydown', this.onKeyDown,
                                    this.agressive);
        browser.removeEventListener('keyup', this.onKeyUp, this.agressive);
        document.getElementById('contentAreaContextMenu').removeEventListener(
            'popupshowing', this.onPopup, this.agressive);
        browser.removeEventListener('contextmenu', this.onContextMenu,
                                 this.agressive);
    },

    /* Unregister some things. */
    shutdown: function(evt) {
        this.agressiveBranch.removeObserver('enabled', this);
        MZPieActions.shutdown(this.prefbranch);
    },

    observe: function(prefbranch, topic, data) {
        if (topic !== 'nsPref:changed')
            throw new Error('huh, I did not sign up for ' + topic);

        if (data !== 'enabled')
            return;

        var browser = getBrowser();
        this.removeEventListeners(browser);
        this.agressive = !this.agressiveBranch.getBoolPref('enabled');
        this.addEventListeners(browser);
    },

    onContextMenu: function(evt) {
        if (this.agressive && this.state !== this.kStateHidden) {
            // We are already displayed or waiting in open-on-drag,
            // and we do not want anything interfering.
            // Especially not silly content-level scripts calling alert()
            // oncontextmenu. So intercept this event:
            evt.preventDefault();
            // HACK: ok, this is fun stuff.
            // Moz's gContextMenu actually gets constructed as a result of
            // this event propagating. If we stop that and then try to
            // manually pop up the context menu it breaks horribly.
            // Since I have not fully tracked down what we are
            // actually stopping here I cannot fix this by simply doing
            // the necessary things here (if I try to construct gContextMenu
            // here it fails because getting document.popupRangeParent raises)
            if (!(this.state === this.kStateOpenOnDrag &&
                  this.prefbranch.getBoolPref('click_for_normal')))
                evt.stopPropagation();
        }
        // afaict evt.defaultPrevented is the DOM3Event standard,
        // getPreventDefault is from nsIDOMNSUIEvent and has the advantage
        // of existing.
        if (!this.agressive && evt.getPreventDefault()) {
            // Someone is trying to prevent a context menu and we are
            // willing to cooperate.
            if (this.state === this.kStateOpenOnDrag)
                this._removeMouseMoveListener();
            else if (this.menu !== null)
                this._hideMenu(false);
        }
    },

    /* Maybe suppress the standard context menu. */
    onPopup: function(evt) {
        /* Showing the normal menu while the radial one is open is
         * always bad. We also block the menu while waiting for drag
         * in drag-on-open.
         *
         * Notice we still block the menu in kStateOpenOnDrag even if
         * the context menu is not on "our" mouse button, but I think
         * that is ok (it is definitely no big deal).
         */
        if (this.state !== this.kStateHidden) {
            evt.preventDefault();
            if (this.agressive)
                evt.stopPropagation();
        }
    },

    /* Change menu entries on alt, cancel menu on esc. */
    onKeyDown: function(evt) {
        if (this.menu !== null) {
            switch (evt.which) {
            case evt.DOM_VK_ALT:
                this.menu.updateEntries(true);
                // On windows hitting alt activates the menu bar, so we
                // need to preventDefault here (TODO: actually test this
                // change on windows)
                evt.preventDefault();
                break;
            case evt.DOM_VK_ESCAPE:
                this._hideMenu(false);
                // not 100% sure preventDefault is sane here, but hey,
                // just hit esc again, we are no longer listening.
                evt.preventDefault();
                break;
            }
        }
    },

    /* Change menu entries on alt. */
    onKeyUp: function(evt) {
        if (this.menu !== null && evt.which === evt.DOM_VK_ALT) {
            this.menu.updateEntries(false);
            // See the DOM_VK_ALT case in onKeyDown
            evt.preventDefault();
        }
    },

    /* Deal with a mouse button being released. */
    onMouseUp: function(evt) {
        /* We only care about other mouse buttons in clicked-open mode.
         * In the other modes it toggles alt state on mouse down, not up.
         * Corner case: we stay in kStateOpenOnDrag when other buttons
         * are clicked.
         */
        if (this.state !== this.kStateClickedOpen &&
            evt.button !== this.prefbranch.getIntPref('mbutton'))
            return;

        switch (this.state) {
        case this.kStateOpenOnDrag:
            // We were not dragged. Stop paying attention.
            this._removeMouseMoveListener();
            if (this.prefbranch.getBoolPref('click_for_normal')) {
                evt.preventDefault();
                // TODO make very sure the coordinates passed here make sense
                document.getElementById('contentAreaContextMenu').showPopup(
                    getBrowser(), evt.clientX, evt.clientY, 'context',
                    null, null);
            }
            /* It is important not to call preventDefault here
             * unless we manually opened the context menu, so
             * things like clicking on links with our menu on the
             * left mouse button work (on some but not all
             * platforms this is the best way to get the context
             * menu open too).
             */
            break;
        case this.kStateShowing:
            if (this.lastMouseX !== evt.clientX ||
                this.lastMouseY !== evt.clientY) {
                // We did not get a mousemove event (or we would be in
                // kStateDragged) but the mouse did move. I think this
                // happens occasionally when the browser and/or system
                // is busy: you click and move the mouse, nothing
                // happens, you release the mouse, and after a while
                // the menu pops up in clicked-open mode. This is
                // annoying (you have to carefully click in the middle
                // to get rid of the menu) so close the menu instead.
                //
                // TODO It might make sense to check the time between
                // the mousedown and mouseup event: if they are close
                // together we can probably accept them (to allow
                // *really* fast back/forward gestures). But that
                // involves figuring out the right magic number for
                // max delay there, which is no fun.
                this._hideMenu(false);
            } else {
                evt.preventDefault();
                this.state = this.kStateClickedOpen;
            }
            break;
        case this.kStateDragged:
        case this.kStateClickedOpen:
            evt.preventDefault();
            this._hideMenu(true);
            break;
        case this.kStateHidden:
            // We get here after the following sequence:
            // - Open the menu in kStateClickedOpen mode.
            // - Press and hold "our" mouse button and another mouse button.
            //   We ignore those events.
            // - Release the one that is not "our" button. We close the menu.
            // - Release "our" button.
            // In this case we want to prevent the popup menu from popping up.
            // (I think this is only really necessary on windows, where
            // the context menu is popped up on mouseup.)

            // Actually, we probably also get here after opening the
            // regular menu through ctrl+rightmouse. Is preventDefault
            // ok then? I think it is, and stuff seems to work, so
            // leave this alone for now.
            evt.preventDefault();
            break;
        default:
            throw new Error('radialcontext bug: unhandled case in onMouseUp');
        }
    },

    /* Deal with mousemove (pop up menu if necessary, move it). */
    onMouseMove: function(evt) {
        /* The check for a coordinate change here is necessary because
         * of mozilla bug 321794: we get a spurious move event when
         * our menu div appears under the mouse cursor.
         */
        if (this.lastMouseX === evt.clientX &&
            this.lastMouseY === evt.clientY)
            return;
        this.lastMouseX = evt.clientX;
        this.lastMouseY = evt.clientY;

        switch (this.state) {
        case this.kStateHidden:
            // Paranoia. Unregister so we do not flood the error message:
            this._removeMouseMoveListener();
            throw new Error(
                'radialcontext bug: mousemove handler not unregistered');
        case this.kStateOpenOnDrag:
            this._showMenu(evt, this.parentNode, this.context);
            this.context = null;
            this.parentNode = null;
            break;
        case this.kStateShowing:
            this.state = this.kStateDragged;
            // fall through
        case this.kStateDragged:
            /* It would be very useful if we could check the current mouse
             * button state here. The button *should* be down in this state,
             * but in some cases the mouse button comes up without us
             * receiving a mouseup event, causing the menu to get "stuck".
             * Unfortunately I cannot find a DOM method or anything
             * moz-specific in the idl files to check the current button
             * state...
             */
            // fall through
        case this.kStateClickedOpen:
            this.menu.update(evt.screenX + this.screenToClientX,
                             evt.screenY + this.screenToClientY);
            this._scheduleTooltip();
            break;
        default:
            throw new Error(
                'radialcontext bug: unhandled case in onMouseMove');
        }
    },

    /* Handle mouse button down: maybe display the menu. */
    onMouseDown: function(evt) {
        // Completely ignore this event if we are being nice and someone
        // is trying to cancel the event (this does not actually work
        // in the limited testing I did, but meh).
        if (!this.agressive && evt.getPreventDefault())
            return;

        // Ignore mousedown here (we activate events in mouseup instead):
        if (this.state === this.kStateClickedOpen)
            return;

        if (evt.button !== this.prefbranch.getIntPref('mbutton')) {
            // Toggle alt-items on alt-click.
            if (this.menu !== null)
                this.menu.updateEntries(!this.menu.altPressed);
            return;
        }

        if (this.menu !== null) {
            // We got a mousedown for "our" button while the menu is
            // being dragged. That should be impossible, but we
            // sometimes miss mouseup events. To make it harder to
            // shoot yourself in the foot by randomly hitting buttons
            // we hide the menu without executing an action here.

            // Ideally we would also do this on mousedown for an alt
            // button, but that is impossible for the same reason
            // we cannot detect this in mousemove: no obvious way to
            // get the current mouse button state. No big deal though,
            // since the alt button cannot do anything destructive.
            this._hideMenu(false);
            return;
        }

        // make sure we got the right button and modifiers
        var mask = (evt.ctrlKey ? 1 : 0) | (evt.shiftKey ? 2 : 0);
        var showMask =
            (this.prefbranch.getBoolPref('show_ctrl') ? 1 : 0) |
            (this.prefbranch.getBoolPref('show_shift') ? 2 : 0);
        var supprMask =
            (this.prefbranch.getBoolPref('suppr_ctrl') ? 1 : 0) |
            (this.prefbranch.getBoolPref('suppr_shift') ? 2 : 0);
        if ((showMask & mask) !== showMask || (
            (supprMask & mask) === supprMask && supprMask !== 0))
            return;

        var targetDoc = evt.originalTarget.ownerDocument;
        // Bail if the browser chrome is targeted.
        if (targetDoc === document)
            return;

        // Check for a usable document type
        const nsIDOMHTMLDocument = Components.interfaces.nsIDOMHTMLDocument;
        const nsIDOMXULDocument = Components.interfaces.nsIDOMXULDocument;
        var parentNode;
        if (targetDoc instanceof nsIDOMHTMLDocument)
            parentNode = targetDoc.body;
        // Bah, this does not actually work. The menu gets added to the tree
        // just fine but is invisible.
        // else if (targetDoc instanceof nsIDOMSVGDocument)
        //     parentNode = targetDoc.documentElement;
        // Making this customizable might be overkill.
        // The rationale is that I like this when working on radialcontext
        // dialogs displayed in browser tabs but others might have
        // more complicated XUL content that needs the right mouse button.
        else if (targetDoc instanceof nsIDOMXULDocument &&
                 this.prefbranch.getBoolPref('show_on_xul'))
            parentNode = targetDoc.documentElement;
        else
            return;

        /* Set up the menu context. Careful: we may still abort (if we
         * find we are in an input element and unwanted) so do not
         * touch "this" yet!
         */

        var selection = evt.view.getSelection();
        /* Somewhat hackish abuse of how javascript objects work here.
         * What we want is a stack of contexts (one per interesting
         * node) that share attributes that do not change per node
         * (like the selection). Javascript's (or perhaps just
         * Mozilla's, not entirely sure) __proto__ attribute lets us
         * do exactly that: anything not explicitly set on the
         * per-node context is looked up on its __proto__ object. The
         * context objects form a linked list (context.next points to
         * the next (less specific) context).
         */
        var globalContext = {
            frameDoc: evt.originalTarget.ownerDocument,
            link: null,
            image: null,
            selection: selection.toString()
        };
        var context = null;
        function chainContext(layout) {
            return {
                layout: layout,
                next: context,
                __proto__: globalContext
            };
        }
        // Start building the context chain at the least specific context:
        context = chainContext('scroll');

        var mouseInSelection = (!selection.isCollapsed &&
                                selection.getRangeAt(0).isPointInRange(
                                    evt.rangeParent, evt.rangeOffset));
        if (context.selection && !mouseInSelection)
            // There is a selection but other context is more important,
            // so add this here at the end of the chain:
            context = chainContext('selection');

        context = chainContext('navigation');

        // XXX this loop is backwards: it adds the deepest (most
        // specific) node at the end (least specific) of the chain,
        // but that is actually useful for the very common
        // image-in-link case (we are more interested in the link than
        // in the image). Would it be better to gather contexts in the
        // correct order and manually push some types to the
        // top/bottom? Are there actual cases other than image-in-link
        // we care about?
        for (var node = evt.originalTarget; node !== null;
             node = node.parentNode) {
            if (node.nodeType === Node.ELEMENT_NODE) {
                switch (node.localName.toLowerCase()) {
                case 'area':
                case 'a':
                    if (node.href) {
                        if (node.href.toLowerCase().indexOf('mailto') === 0)
                            context = chainContext('mail');
                        else
                            context = chainContext('link');
                        context.link = node;
                    }
                    break;
                case 'img':
                    context = chainContext('image');
                    context.image = node;
                    break;
                case 'input':
                    // Don't hinder working with input fields:
                    if (evt.button === 0)
                        return;
                    if (node.type.toLowerCase() !== 'text' &&
                        node.type.toLowerCase() !== 'password')
                        break;
                    // Fall through, treat as a textarea.
                case 'textarea':
                    // Don't hinder working with input fields:
                    if (evt.button === 0)
                        return;
                    context = chainContext('edit');
                    break;
                }
            }
        }

        if (mouseInSelection)
            // Primary context is selection.
            context = chainContext('selection');

        // If we got this far we are really going to respond to this event.

        this.lastMouseX = evt.clientX;
        this.lastMouseY = evt.clientY;

        /* So where do we listen for events?
         *
         * If we listen on the main content document we do not receive
         * events if the mouse is over an iframe *unless* a node
         * "belonging" to the main document is displayed on top of the
         * iframe. Our menu is such an element. This means the menu
         * gets "stuck" if you move the mouse *quickly* into the
         * iframe.
         *
         * If we listen on getBrowser() we keep getting the events,
         * but evt.client{X,Y} "jumps" when the mouse moves to a
         * different document. I cannot find an api to convert the
         * coordinates, so instead we use screen coordinates and just
         * remember the offset for this first event and use that to
         * correct future events.
         *
         * TODO: when the mousedown has an iframe as target it would be
         * even better to attach the menu to the parent document instead
         * of the iframe, but I do not know how to determine the screen ->
         * client coordinate offsets in that case.
         */
        this.screenToClientX = evt.clientX - evt.screenX;
        this.screenToClientY = evt.clientY - evt.screenY;
        this.mouseMoveElement = getBrowser();
        this.mouseMoveElement.addEventListener(
            'mousemove', this.onMouseMove, true);

        /* We stop the contextmenu in a different way because mozilla's
         * "do not allow js to mess with the contextmenu" pref makes this
         * not work. But we still need this when we are not using the
         * right mouse button.
         */
        evt.preventDefault();
        if (this.agressive)
            evt.stopPropagation();

        if (this.prefbranch.getBoolPref('dragonly')) {
            this.context = context;
            this.parentNode = parentNode;
            this.state = this.kStateOpenOnDrag;
        } else
            this._showMenu(evt, parentNode, context);
    }
};


// Wrap the functions to get "this" right. Remember to wrap this in a
// function if this is converted back to <1.7 js to prevent leaking
// the loop variable(s).
for (let [name, value] in Iterator(MZPieIntegration)) {
    /* Ok, so this is pretty special. Assuming I parse the bug
     * comments correctly: since this is chrome code its functions get
     * precompiled once, meaning the functions above have [object
     * nsXULPrototypeScript compilation scope].Function as their
     * class/constructor. At the time this loop runs the global object
     * is the XUL window, and window.Function is not identical to the
     * Function that constructed those precompiled functions. So this
     * *has* to be "typeof(value) === 'function'", "value instanceof
     * Function" fails because it does not use the same Function
     * object. Bug 300079, 321231.
     *
     * This is fixed in Firefox 3, so we can use "value instanceof
     * Function" once we require that version.
     */
    if (typeof(value) === 'function') {
        // You can do this without the "let" by using a separate
        // factory function that generates the wrapper.
        let original = value;
        MZPieIntegration[name] = function(/* *args */) {
            return original.apply(MZPieIntegration, arguments);
        };
    }
}


window.addEventListener('load', MZPieIntegration.init, true);
