/*
    Copyright (c) 2002 Jens Tinz
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* This file implements the actual radial menu.
 *
 * It can be attached to HTML documents and updated in response to
 * events. There are only class definitions here, no global state and
 * no code registering/unregistering event handlers. The code should
 * not depend on running in a browser (using functions from browser.js).
 *
 * MZPieMenu is the only toplevel name defined.
 */

/* Class controlling a displayed menu.
 *
 * parentNode is the node to add the menu to (normally the html "body").
 * layout is an array of MZPieMenu.Actions.
 * mx and my are the initial position in pixels (client area-relative).
 * altPressed is a bool, true to start out with the alt set of actions.
 * keepLabels is a bool, true to keep the tooltip up when moving.
 * zoom is a float used for adjusting tooltip text size.
 * context is an arbitrary object stored as this.context, used by actions.
 */
function MZPieMenu(parentNode, layout, mx, my, altPressed, keepLabels,
                   textZoom, context) {
    // Apparently needed to be subclassable. Isn't js fun!
    if (!arguments.length)
        return;

    this.x = mx;
    this.y = my;
    this.altPressed = altPressed;
    this.keepLabels = keepLabels;
    this.textZoom = textZoom;
    this.context = context;

    // index of item under mouse (or null when the mouse is in the hole)
    this._sector = null;
    // root of DOM tree containing the menu labels
    this._lnode = null;
    // label nodes
    this._normalLabels = this._boldLabels = null;

    /* Create a div to contain the items. This is used as just a
     * coordinate system, not a "box", with the origin at the menu's
     * center. The rest of the menu is image nodes added to this div.
     * The tooltip is inside a second div, added as a child of this
     * one in showLabels.
     */
    // HACK: _createNode grabs the document from this.node:
    this.node = parentNode;
    this.node = this._createNode('div', 'fixed', mx + 'px', my + 'px',
                                 '0', 'auto');
    this.node.style.zIndex = '32767';

    // add the menu image
    var timg = this._createNode(
        'img', 'relative',
        '-' + this.menuXOffset + 'px', '-' + this.menuYOffset + 'px',
        'auto', 'auto');
    timg.src = this.path + 'menu.png';
    timg.alt = '';
    this.node.appendChild(timg);

    // The icon positions are calculated once and stored here:
    this._iconX = [];
    this._iconY = [];
    // Array of action icons (no null entries):
    this._icons = [];
    // Array of "has alt action" indicators (with null entries):
    this._alticons = new Array(layout.length);
    // Array of "has submenu" indicators (with null entries):
    this._subicons = new Array(layout.length);

    for (var i = 0; i < layout.length; ++i) {
        var angle = ((i + 1) / layout.length) * 2 * Math.PI;
        this._iconX.push(-this.iconXOffset + Math.round(
            this.imagesRadius * Math.sin(angle)));
        this._iconY.push(-this.iconYOffset - Math.round(
            this.imagesRadius * Math.cos(angle)));
        this._icons.push(this._createIcon(null, i));
        // Slight hack: src is filled out by applyLayout.
    }

    this.applyLayout(layout);

    // Do this last so the menu appears in the right location.
    parentNode.appendChild(this.node);
}

MZPieMenu.prototype = {
    // Path to icons base directory (trailing '/' is required).
    path: 'chrome://radialcontext-mz/skin/',
    // Coordinates of the menu center in menu.png.
    menuXOffset: 46,
    menuYOffset: 46,
    // Radius of the hole, no selection until mouse is outside this.
    holeRadius: 17,
    // Distance from menu center to center of action icon.
    imagesRadius: 30,
    // Distance at which the menu starts to move or a submenu opens.
    actionRadius: 36,
    // Coordinates of the center of an icon.
    iconXOffset: 11,
    iconYOffset: 11,
    // Coordinates of the center of labels.png.
    labelsMainXOffset: 52,
    labelsMainYOffset: 55,
    // Distance from the top of the image to the top of the first label.
    labelsLeftYOffset: 1,
    labelsRightYOffset: 23,
    // Height of a single label.
    labelHeight: 22,

    /* Show a different array of Actions. */
    applyLayout: function(layout) {
        this.normalentries = layout.map(function(newitem, i, layout) {
            var item;
            // tweak returns null or undefined when done, so no !== here
            while (newitem != null) {
                item = newitem;
                newitem = item.tweak(this.context);
            }
            return item;
        }, this);
        this.altentries = this.normalentries.map(
            function(normalitem, i, normalentries) {
                var item = normalitem;
                var newitem = item.altitem;
                // tweak returns null or undefined when done, so no !== here
                while (newitem != null) {
                    item = newitem;
                    newitem = item.tweak(this.context);
                }
                // show/hide the alt indicator if necessary
                if (normalitem === item) {
                    if (this._alticons[i] != null) {
                        this.node.removeChild(this._alticons[i]);
                        this._alticons[i] = null;
                    }
                } else if (this._alticons[i] == null)
                    this._alticons[i] = this._createIcon(
                        'alt_indicator.png', i);
                return item;
            }, this);

        this.updateEntries(this.altPressed);
    },

    /* Changes the entries according to status of alt modifier.
     *
     * Also called internally by applyLayout.
     */
    updateEntries: function(altPressed) {
        this.altPressed = altPressed;
        if (altPressed)
            this.entries = this.altentries;
        else
            this.entries = this.normalentries;

        this.entries.forEach(function(entry, i, entries) {
            this._icons[i].src = this.path + entry.iconpath;

            // Show/hide submenu indicators if necessary:
            if (entry instanceof MZPieMenu.Submenu) {
                // Initialized to undefined, not null, so "=="
                if (this._subicons[i] == null) {
                    var name = [
                        'submenu_indicator_ne.png', null,
                        'submenu_indicator_se.png', null,
                        'submenu_indicator_sw.png', null,
                        'submenu_indicator_nw.png', null,
                    ][entries.length > 4 ? i : i * 2];
                    if (name != null)
                        this._subicons[i] = this._createIcon(name, i);
                }
            } else if (this._subicons[i] != null) {
                this.node.removeChild(this._subicons[i]);
                this._subicons[i] = null;
            }

            // Update labels if we are displaying them:
            if (this._lnode !== null) {
                this._normalLabels[i].textContent = entry.label;
                this._boldLabels[i].textContent = entry.label;
            }
        }, this);
    },

    /* Internal helper for creating icon nodes. */
    _createIcon: function(path, index) {
        var timg = this._createNode(
            'img', 'absolute',
            this._iconX[index] + 'px', this._iconY[index] + 'px',
            'auto', 'auto');
        if (path !== null)
            timg.src = this.path + path;
        timg.alt = '';
        this.node.appendChild(timg);
        return timg;
    },

    /* Remove the menu.
     *
     * If doAction is true also execute the currently selected action.
     *
     * Calling any other function on the menu after this might break.
     */
    close: function(doAction) {
        /* We call doAction last because some actions ("save as"
         * specifically) apparently block, and it looks weird if they
         * block with the menu open.
         */
        this.node.parentNode.removeChild(this.node);
        if (doAction && this._sector !== null) {
            var entry = this.entries[this._sector];
            // "Activates" non-action items too, but that's a noop so it's ok.
            if (entry.active)
                entry.doAction(this.context);
        }
    },

    /* Update mouse position. */
    update: function(mx, my) {
        // identify current sector
        var nsec = null;
        var dx = mx - this.x;
        var dy = my - this.y;
        var dst2 = dx * dx + dy * dy;
        if (dst2 > this.holeRadius * this.holeRadius) {
            // Figure out which half-sector we are in.
            // the - 1 makes the first half of entries[0] halfsector 0.
            var hsec = Math.floor(
                this.entries.length * Math.atan2(dx, -dy) / Math.PI) - 1;
            hsec %= this.entries.length * 2;
            if (hsec < 0)
                hsec += this.entries.length * 2;
            // Slightly silly way to do integer division by 2
            // (silly js does not have an integer division operator).
            nsec = hsec >> 1;
            // If the target is a dummy adjust to the nearest neighbour.

            // We *only* do this for actual dummy items, not for
            // inactive items, or the "forward" nav action being
            // inactive makes the tab and page submenus too "sensitive".
            if (this.entries[nsec] instanceof MZPieMenu.Dummy) {
                // First or second half of the dummy?
                if (hsec % 2)
                    // Second half.
                    ++nsec;
                else
                    // First half. Make sure nsec stays positive.
                    nsec += this.entries.length - 1;
                nsec %= this.entries.length;
            }

            if (dst2 > this.actionRadius * this.actionRadius) {
                var item = this.entries[nsec];
                if (item instanceof MZPieMenu.Submenu) {
                    // Show submenu.
                    /* The order here matters: ContextMenu changes
                     * this.context here, and that may affect tweak functions
                     * triggered by applyLayout.
                     */
                    item.activateMenu(this);
                    this.applyLayout(item.layout);
                    this._sector = null;
                    this.x = mx;
                    this.y = my;
                } else {
                    // drag menu
                    if (!this.keepLabels)
                        this.hideLabels();

                    var fak = this.actionRadius / Math.sqrt(dst2);
                    this.x = mx - Math.round(dx * fak);
                    this.y = my - Math.round(dy * fak);
                }
                this.node.style.left = this.x + 'px';
                this.node.style.top = this.y + 'px';
            }
        }

        // Update current sector. Keep the labels visibility code
        // synced with showLabels.
        if (nsec !== this._sector) {
            // Deactivate previous rollover if there was one:
            if (this._sector !== null) {
                this._icons[this._sector].style.opacity = '1.0';
                if (this._lnode !== null) {
                    this._boldLabels[this._sector].style.visibility =
                        'hidden';
                    this._normalLabels[this._sector].style.visibility =
                        'visible';
                }
            }
            // Activate new rollover if there is one:
            if (nsec !== null && this.entries[nsec].active) {
                this._icons[nsec].style.opacity = '0.64';
                if (this._lnode !== null) {
                    this._normalLabels[nsec].style.visibility = 'hidden';
                    this._boldLabels[nsec].style.visibility = 'visible';
                }
            }
            this._sector = nsec;
        }
    },

    /* Show the tooltip with labels.
     *
     * Calling this if it is already displayed is safe.
     */
    showLabels: function() {
        // create resources if necessary
        if (this._lnode === null) {
            /* Main container floating behind the menu, with the
             * origin in the top left of the center image. Size is
             * picked up from the image: height so we can size the
             * label containers based on it, width so we can position
             * the righthand labels based on it.
             */
            var main = this._createNode(
                'div', 'absolute',
                '-' + this.labelsMainXOffset + 'px',
                '-' + this.labelsMainYOffset + 'px',
                'auto', 'auto');
            main.style.zIndex = '-1';

            // Center image.
            var timg = this._createNode(
                'img', 'static', '0', '0', 'auto', 'auto');
            timg.src = this.path + 'labels_center.png';
            timg.alt = '';
            main.appendChild(timg);

            /* If we simply change the font weight of the labels
             * dynamically we trigger a reflow of the tooltip which
             * looks a bit ugly. So instead we use the bold labels to
             * size the background image, add the normal labels to a
             * second (transparent) div on top of that, then toggle
             * the visibility of the labels.
             *
             * Notice we still resize the label when alt is
             * pressed/released. We could avoid that too by adding
             * some more hidden labels but that is probably a bad idea
             * since their sizes may differ a lot ("Reload" vs "Reload
             * - Skip Cache"...), so a resize is better than a label
             * with a silly size.
             */

            /* Containers for the labels. Positioned next to the main
             * div. Height forced to the height of the parent to make
             * sure the entire background image shows up. Width
             * determined by the actual labels inside.
             */
            var leftBold = this._createNode(
                'div', 'absolute', 'auto', '0', 'auto', '100%');
            leftBold.style.right = '100%';
            var leftNormal = this._createNode(
                'div', 'absolute', '0', '0', 'auto', '100%');
            leftBold.style.backgroundImage = (
                'url("' + this.path + 'labels_left_pad.png")');
            var rightBold = this._createNode(
                'div', 'absolute', '100%', '0', 'auto', '100%');
            var rightNormal = this._createNode(
                'div', 'absolute', '0', '0', 'auto', '100%');
            rightBold.style.backgroundImage = (
                'url("' + this.path + 'labels_right_pad.png")');
            // Make sure background css props are sane for both divs.
            [leftBold, rightBold].forEach(function(node, i, nodes) {
                node.style.backgroundAttachment = 'scroll';
                node.style.backgroundPosition = 'top left';
                node.style.backgroundRepeat = 'repeat-x';
            });
            main.appendChild(leftBold);
            main.appendChild(rightBold);
            leftBold.appendChild(leftNormal);
            rightBold.appendChild(rightNormal);

            // Attach the left and right edge images to the left and
            // right divs the same way those are attached to the main div:
            var node = this._createNode(
                'img', 'absolute', 'auto', '0', 'auto', 'auto');
            node.style.right = '100%';
            node.src = this.path + 'labels_left_edge.png';
            leftBold.appendChild(node);
            node = this._createNode('img', 'absolute', '100%', '0',
                                    'auto', 'auto');
            node.src = this.path + 'labels_right_edge.png';
            rightBold.appendChild(node);

            /* The divs containing the actual label text use relative
             * positioning to put them on the right height. Each label
             * is forced to stay on one line, so their width
             * determines the width of their containing div (which
             * makes the background the right size), and the labels
             * below show up at the right height.
             */
            // ">> 1" is a silly way to do integer division by 2
            var rightEntries = this.entries.slice(0,
                                                  this.entries.length >> 1);
            var boldLabels = rightEntries.map(function(entry, i, entries) {
                    return this._createLabel(this.labelsRightYOffset,
                                             entry.label, rightBold, true);
                }, this);
            var normalLabels = rightEntries.map(function(entry, i, entries) {
                    return this._createLabel(this.labelsRightYOffset,
                                             entry.label, rightNormal, false);
                }, this);

            // We need to add those to the container reversed.
            var leftEntries = this.entries.slice(this.entries.length >> 1);
            leftEntries.reverse();
            var leftLabels = leftEntries.map(function(entry, i, entries) {
                    return this._createLabel(this.labelsLeftYOffset,
                                             entry.label, leftBold, true);
                }, this);
            leftLabels.reverse();
            this._boldLabels = boldLabels.concat(leftLabels);

            leftLabels = leftEntries.map(function(entry, i, entries) {
                    return this._createLabel(this.labelsLeftYOffset,
                                             entry.label, leftNormal, false);
                }, this);
            leftLabels.reverse();
            this._normalLabels = normalLabels.concat(leftLabels);

            // hide/unhide labels (keep synced with update):
            for (var i = 0; i < this.entries.length; ++i) {
                var active = (i === this._sector &&
                              this.entries[this._sector].active);
                this._boldLabels[i].style.visibility = (
                    active ? 'visible' : 'hidden');
                this._normalLabels[i].style.visibility = (
                    active ? 'hidden' : 'visible');
            }

            // Do this last so it appears in the right spot.
            this.node.appendChild(main);
            this._lnode = main;
        }

        this._lnode.style.display = 'block';
    },

    /* Helper used by showLabels. */
    _createLabel: function(ypos, label, container, bold) {
        var tdiv = this._createNode(
            'div', 'relative', '0', ypos + 'px',
            'auto', this.labelHeight + 'px');

        // TODO: check if there is a way to specify fontsize in px
        // independent of the text zoom value (is it a moz bug text
        // zoom is applied here?)

        // Same font value used for actual popups (see popup.css).
        tdiv.style.font = 'message-box';

        /* Compensate text zoom. Doing this here means we do not
         * adjust if zoom changes between tooltip displays without
         * reopening the menu, but I can live with that.
         */
        // XXX labelHeight - 2 is "magic".
        tdiv.style.lineHeight = Math.round(
            (this.labelHeight - 2) / this.textZoom) + 'px';
        tdiv.style.fontSize = Math.round(8 / this.textZoom) + 'pt';

        tdiv.style.fontWeight = bold ? 'bold' : 'normal';

        // _createNode-style cleanup for text.
        tdiv.style.letterSpacing = 'normal';
        tdiv.style.verticalAlign = 'middle';
        tdiv.style.textAlign = 'left';
        // textDecoration and textIndent is set in _createNode.
        tdiv.style.textTransform = 'none';
        // Important! without this moz tries to make every
        // label as narrow as possible.
        tdiv.style.whiteSpace = 'nowrap';
        tdiv.style.wordSpacing = 'normal';

        tdiv.textContent = label;

        container.appendChild(tdiv);
        return tdiv;
    },

    /* Hide the tooltip with labels if currently visible.
     *
     * Calling this while the labels are already hidden is safe.
     */
    hideLabels: function() {
        if (this._lnode !== null)
            this._lnode.style.display = 'none';
    },

    /* Internal helper undoing most style rules on a new element.
     *
     * Arguments are the element type and required css properties.
     * If the element will contain text this function is not sufficient.
     */
    _createNode: function(nodeName, position, left, top, width, height) {
        var node = this.node.ownerDocument.createElementNS(
            'http://www.w3.org/1999/xhtml', nodeName);
        node.style.position = position;
        node.style.left = left;
        node.style.top = top;
        node.style.width = width;
        node.style.height = height;

        // Sorted based on nsIDOMCSS2Properties.idl. All these are
        // here because a test page setting them on all div and img
        // elements affected the menu, so do not remove stuff here
        // without testing the menu (including tooltip).
        node.style.backgroundColor = 'transparent';
        node.style.backgroundImage = 'none';
        node.style.border = 'none';
        node.style.bottom = 'auto';
        // TODO: "clear"?
        node.style.clip = 'auto';
        node.style.color = 'black';
        // XXX div:before { content: 'blah' } breaks stuff.
        node.style.cursor = 'auto';
        node.style.direction = 'ltr';
        node.style.display = 'block';
        node.style.cssFloat = 'none';
        node.style.margin = '0';
        node.style.maxHeight = 'none';
        node.style.maxWidth = 'none';
        node.style.minHeight = '0';
        node.style.minWidth = '0';
        node.style.outlineStyle = 'none';
        node.style.overflow = 'visible';
        node.style.padding = '0';
        node.style.right = 'auto';
        // Just setting this on nodes containing text is apparently
        // not sufficient.
        node.style.textDecoration = 'none';
        node.style.textIndent = '0';
        node.style.verticalAlign = 'baseline';
        node.style.visibility = 'visible';

        node.style.opacity = '1';

        return node;
    }
};


/* Basic action. Initialize with a label string and path to an icon.
 *
 * Override doAction (subclass or reassign after instantiation) to
 * make it actually do something.
 */
MZPieMenu.Action = function(label, iconpath) {
    this.label = label;
    this.iconpath = iconpath;
    this.altitem = null;
    this.active = true;
};

MZPieMenu.Action.prototype = {
    /* Called before this item is displayed.
     *
     * Use this for context-sensitive actions or menus. It can either
     * change this item's attributes or return a different item.
     *
     * Remember this still gets called for menu previews and the like
     * where doAction will never be called, so should make fewer
     * assumptions (specifically it should not depend on browser.js).
     *
     * Tweak functions are called recursively, return null or
     * undefined when done (returning "this" triggers an infinite
     * loop).
     */
    tweak: function(context) {
        return null;
    },

    /* Actually execute this action. */
    doAction: function(context) {
    }
};


/* Special kind of action for submenus. layout is an array of Actions. */
MZPieMenu.Submenu = function(label, iconsrc, layout) {
    MZPieMenu.Action.call(this, label, iconsrc);
    this.layout = layout;
};

MZPieMenu.Submenu.prototype = new MZPieMenu.Action;

/* Called right before this submenu is displayed.
 *
 * Does not have a lot of users. This interface may change.
 */
MZPieMenu.Submenu.prototype.activateMenu = function(menu) {};


/* A "hole" in the menu. Special because it passes focus to its neighbours. */
MZPieMenu.Dummy = function() {
    MZPieMenu.Action.call(this, '', 'dummy.png');
    this.active = false;
};

MZPieMenu.Dummy.prototype = new MZPieMenu.Action;
